package de.korap.services;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;

import de.korap.services.LemmaResponse;

public class LemmaInfo {
	
	// GLEMM API Options:
	final static String
		GLEMM_OPTS_FLEX 	= "flex",	// Flexionsformen
		GLEMM_OPTS_COMP		= "comp", 	// Komposita
		GLEMM_OPTS_OTHER	= "other",	// sonstige Wortbildungsformen
		GLEMM_OPTS_SPEC		= "spec",	// Spezialfälle.
		GLEMM_OPTS_NFLEX 	= "-flex",	// keine Flexionsformen
		GLEMM_OPTS_NCOMP	= "-comp", 	// keine Komposita
		GLEMM_OPTS_NOTHER	= "-other",	// keine sonstige Wortbildungsformen
		GLEMM_OPTS_NSPEC	= "-spec";	// keine Spezialfälle.
	
	// Options Settings requested in Query:
	final static int
		OPTS_SETTING_UNDEF	= 0,
		OPTS_SETTING_C2		= 1,	// lemma/Schuh and lemma/Schuh?opts=comp...
		OPTS_SETTING_INT	= 2;	// lemma/Schuh?prop1=... [internal mode].
	
	public String
		lemma,			// requested lemma, after conversion = 'ß' -> 'ss', i.e. 'schliesslich'.
		lemmaOrig;		// original lemma version, as submitted (e.g. 'schließlich').
	public String
		lemmaOpts;		// lemmatization options as written in request.
	public int
		prop1		= 2, 
		prop2		= 2,	// lemmatisation rule: 0 = '-', 1 = '+', 2 = undef.
		spec		= 2;
	// Options set by user: oFlex, oComp etc.
	// oFlex = 0 : no flexions / 1 : flexions / 2 : undefined.
	public int
		oFlex		= 2,
		oComp		= 2,
		oOther		= 2,
		oSpec		= 2;
	public int
		optsSetting = OPTS_SETTING_C2;
	Connection 
		con = null;
	Statement
		stmt = null;	
	public ResultSet
		res = null;
	public int
		nWfs = -1; 	// -1: error occured; 0 = lemma has no wordforms; > 0 = no. of
					//     wordforms extracted for that lemma.
					// nWfs is filled after this.res has be read.
	StringBuilder
		wordforms = null;


	/* lemmaInfo
	 * 
	 * constructor.
	 * 13.01.20/FB
	 */
	LemmaInfo()
	
	{
	lemma 		= null;
	lemmaOrig	= null;
	prop1 		= 2;
	prop2 		= 2;
	}

	/* lemmaInfo(lemma)
	 * 
	 * constructor.
	 * 13.01.20/FB
	 */
	LemmaInfo(final String lemma)
	
	{
	this.lemma 		= lemma.replace("ß",  "ss");
	this.lemmaOrig 	= lemma;
	this.prop1 		= 2;
	this.prop2 		= 2;
	}
	
	/* LemmaInfo(lemma, prop1, prop2)
	 * 
	 * construtor.
	 * E.g. "Haus&p1=1&p2=0"
	 * To be replaced.
	 * 13.01.20/FB
	 */
	
	LemmaInfo(String lemma, int prop1, int prop2)
	
	{
	this.lemma 		= lemma.replace("ß",  "ss");
	this.lemmaOrig 	= lemma;
	this.prop1 		= prop1 == 0 || prop1 == 1 ? prop1 : 2;
	this.prop2 		= prop2 == 0 || prop2 == 1 ? prop2 : 2;
	}
	
	/* parseAllOptions
	 * 
	 * Input: options of C2 Option Setting of options of Extended Options Setting
	 *        through Query Params.
	 *        Either C2 Setting: opts, or
	 *        Extended Setting : prop1, prop2, spec
	 * will be set.
	 * If options = "undef", extended Options are requested instead of C2 Options.
	 * If prop1 == prop2 == spec == 3: Extended Setting is not requested.
	 * Default, if both are not set: options=flex.
	 * 
	 * Parameter options is parsed, checked against correct constants and
	 * stored into lemmaInfo.GlemmXXX.
	 * 
	 * Accepted constants: GLEMM_OPTS_x
	 * E.g. "flex+comp"
	 * Note: 
	 * - GlemmFlex is set either explicitly by sumitting option "flex", or
	 *   implicitly by submitting any other option.
	 * - First of all, we implement the C2 Option Setting.
	 * 15.01.20/FB
	 * - implementation of extended flag options: 31.01.20/FB
	 */
	
	int parseAllOptions(String options, int prop1, int prop2, int spec)
	
	{
	if( !options.equals("undef") )
		{
		this.optsSetting = OPTS_SETTING_C2;
		return parseOpts_C2Setting(options);
		}
	else if( prop1 != 3 || prop2 != 3 || spec != 3)
		{
		this.optsSetting = OPTS_SETTING_INT;
		return parseOpts_ExtendedSetting(prop1, prop2, spec);
		}
	else
		{	// no setting requested, return default
		this.optsSetting = OPTS_SETTING_C2;
		return parseOpts_C2Setting("flex");	
		}
		
	} // parseAllOptions

	/*
	 *  parseOpts_C2Setting
	 *  
	 *  expected: "(flex) (comp) (other) (spec)"
	 *  Notes:
	 *  - Jersey transforms "flex+comp+other..." automatically to
	 *    "flex comp other ...".
	 *    
	 *  31.01.20/FB
	 */
	
	int parseOpts_C2Setting(final String options)
	
	{
	final String
		func = "parseOpts_C2Setting";
	String[]
			opts = options.split("[ ]");
		int
			nErrors = 0;
		
	this.prop1 = this.prop2 = 0; // default: +Flex -Comp -Other -Spec.
	this.oFlex = 1;				 // default: +Flex
	
	for(int i=0; i<opts.length; i++)
		{
		switch(opts[i]) {
		
		case GLEMM_OPTS_FLEX: 
			this.oFlex	= 1;
			break;
		case GLEMM_OPTS_COMP: 
			this.oComp	= 1;
			break;
		case GLEMM_OPTS_OTHER: 
			this.oOther = 1;
			break;
		case GLEMM_OPTS_SPEC: 
			this.oSpec = 1;
			break;
		default:
			if( opts[i].isEmpty() )
				{	
				this.oFlex = 1;
				}
			else
				{
				GlemmServices.ferr.printf("Error: %s: unknown Glemm Option '%s'!\n", func, opts[i]);
				nErrors++;
				}
			} // switch
		}
	
	if( nErrors > 0 )
		return nErrors;
	
	// convertion of User Options -> internal flags:

	this.prop1 = this.prop2 = this.spec = 0;
	
	// prop1 = 2 = undef = Flex+Comp requested; 1 = only Comp requested.
	if( this.oComp == 1 )
		this.prop1 = 2; // undef = flex and comp.
	
	if( this.oOther == 1 )
		this.prop2 = 2; // undef: other and not other requested.
	
	if( this.oSpec == 1 )
		this.spec = 2; // undef: spec and not spec requested.

	return 0; // OK
	} // parseOptions_C2Setting

	/*
	 *  parseOpts_ExtendedSetting
	 *  
	 *  expected: prop1, prop2, spec may have the following values:
	 *  		  0 : false;
	 *  		  1 : true;
	 *  		  2 : undef, i.e. do not set that flag for SQL Request.
	 *  		  3 : flag not requested
	 *  Notes:
	 *  - flags are directly set by the sumitted parameters.
	 *  31.01.20/FB
	 */
	
	int parseOpts_ExtendedSetting(final int prop1, final int prop2, final int spec)
	
	{
	if( prop1 == 0 || prop1 == 1 )
		this.prop1 = prop1;
	else
		this.prop1 = 2; // undef
	
	if( prop2 == 0 || prop2 == 1 )
		this.prop2 = prop2;
	else
		this.prop2 = 2; // undef

	if( spec == 0 || spec == 1 )
		this.spec = spec;
	else
		this.spec = 2; // undef

	return 0; // OK
	
	} // parseOptions_ExtendedSetting
	
	/* parseOptions
	 * 
	 * @Deprecated
	 * 
	 * Input: Path Param options.
	 * Parameter options is parsed, checked against correct constants and
	 * stored into lemmaInfo.GlemmXXX.
	 * 
	 * Accepted constants: GLEMM_OPTS_x
	 * E.g. "flex+comp"
	 * Note: 
	 * - GlemmFlex is set either explicitly by sumitting option "flex", or
	 *   implicitly by submitting any other option.
	 * - First of all, we implement the C2 Option Setting.
	 * 15.01.20/FB
	 */
	
	int parseOptions(String options)
	
	{
	final String
		func = "parseOptions"; 
	final boolean
		modeC2 = true;	// COSMAS II Option Setting.
	String[]
		opts = options.split("[ ,]");
	int
		nErrors = 0;
	
	this.prop1 = this.prop2 = 0; // default: +Flex -Comp -Other -Spec.
	
	for(int i=0; i<opts.length; i++)
		{
		//System.out.printf("Debug: %s: i=%d opts='%s'.\n", func, i, opts[i]); 
		switch(opts[i]) {
		
		case GLEMM_OPTS_FLEX: 
			this.oFlex	   = 1;
			break;
		case GLEMM_OPTS_NFLEX: 
			this.oFlex	   = 0;
			break;
		case GLEMM_OPTS_COMP: 
			this.prop1 = 1;
			this.oComp = 1;
			break;
		case GLEMM_OPTS_NCOMP: 
			this.prop1 = 1;
			this.oComp = 0;
			break;
		case GLEMM_OPTS_OTHER: 
			this.prop2 = 1;
			this.oOther = 1;
			break;
		case GLEMM_OPTS_NOTHER: 
			this.prop2 = 1;
			this.oOther = 0;
			break;		
		case GLEMM_OPTS_SPEC: 
			// neither of prop1 nor prop2 is set by this option.
			this.oSpec = 1;
			break;
		case GLEMM_OPTS_NSPEC: 
			// neither of prop1 nor prop2 is set by this option.
			this.oSpec = 0;
			break;
		default:
			if( opts[i].isEmpty() )
				{	
				// don't set anything here.
				}
			else
				{
				GlemmServices.ferr.printf("Error: %s: unknown GLEMM Option '%s'!\n", func, opts[i]);
				nErrors++;
				}
			} // switch
		}
	
	if( nErrors > 0 )
		return nErrors;
	
	// convertion of User Options -> internal flags:
	// COSMAS II Options Setting:
	if( modeC2 )
		{
		this.prop1 = this.prop2 = this.spec = 0;
		
		// prop1 = 2 = undef = Flex+Comp requested; 1 = only Comp requested.
		if( this.oComp == 1 )
			this.prop1 = this.oFlex == 1 ? 2 : 1; 
		
		if( this.oOther == 1 )
			this.prop2 = 2; // undef: other and not other requested.
		
		if( this.oSpec == 1 )
			this.spec = 2; // undef: spec and not spec requested.
		}
	
	return nErrors; // OK
	} // parseOptions
	
	/* optionstoString:
	 * 
	 * 15.01.20/FB
	 */
	
	String optionstoString()
	
	{
	StringBuilder
		opts = new StringBuilder("");
	int
		n = 0;
	
	if( this.oFlex != 2 )
		{
		opts.append( (this.oFlex == 0 ? "-" : "") + GLEMM_OPTS_FLEX);
		n++;
		}
	if( this.oComp != 2 )
		{
		opts.append( (this.oComp == 0 ? "-" : n > 0 ? "+" : " ") + GLEMM_OPTS_COMP);
		n++;
		}
	if( this.oOther != 2 )
		{
		opts.append( (this.oOther == 0 ? "-" : n > 0 ? "+" : "") + GLEMM_OPTS_OTHER);
		n++;
		}
	if( this.oSpec != 2 )
		{
		opts.append((this.oSpec == 0 ? "-" : n > 0 ? "+" : "") + GLEMM_OPTS_SPEC);
		}
	
	return opts.toString();
	} // optionstoString
	
	/* * * * * * * * * * * * *
	 * close:
	 * 
	 * closes the SQL stuff used for extracting the lemma's wordforms.
	 * 14.01.20/FB
	 * * * * * * * * * * * * */
	
	void close()
	
	{
	GlemmServices.fout.printf("Debug: close: res=%s con=%s stmt=%s.\n",
				this.res != null ? "allocated" : "null",
				this.con != null ? "allocated" : "null",
				this.stmt!= null ? "allocated" : "null");
		
	try {
		if( this.res != null )
			this.res.close();
		if( this.con != null )
			this.con.close();
		if( this.stmt != null )
			this.stmt.close();
		this.res 	= null;
		this.con 	= null;
		this.stmt 	= null;
		} 
	catch (SQLException e) {
		e.printStackTrace();
		}
	
	} // close

	/* wfs2HTML
	 * writes wfs to a HTML formatted buffer.
	 * 14.01.20/FB
	 */
	
	public String wfs2HTML()
	
	{
	int
		n = -1; // undefined.

	this.wordforms = new StringBuilder();
	
	try {
		for(n=0; this.res.next(); n++)
			{
			this.wordforms.append(String.format("<tr><td>%s</td></tr>\n",  this.res.getString("wf")));
			}
		} 
	catch (SQLException e) {
		n = -1; // reset to undefined due to error.
		e.printStackTrace();
		}

	this.nWfs = n;
	
	return this.wordforms != null ? this.wordforms.toString() : null;
	
	} // wfs2HTML

	/* res2JSON():
	 * 
	 * transforms result of Lemma Query to a JSON list.
	 * 27.01.20/FB
	 */
	
	public LemmaResponse res2JSON()
	
	{
	final String
		func = "res2JSON";
	LemmaResponse
		lemResp = null;
	
	try {
		lemResp = new LemmaResponse();
		lemResp.listofWfs = new ArrayList<>();
		}
	catch ( Exception e ){
		GlemmServices.ferr.printf("Error: %s: cannot allocate memory for %s!\n",
				func, lemResp == null ? "lemResp" : "listofWfs");
		e.printStackTrace();
		if( lemResp == null )
			return null;
		else
			return lemResp;
		}
	
	// fill header:
	lemResp.head_nWfs    = -1;
	lemResp.head_query   = this.lemmaOrig;
	lemResp.head_options = this.optionstoString();
	
	// fill lemma list:
	
	try {
		for(this.nWfs=0; this.res.next(); this.nWfs++)
			{
			lemResp.listofWfs.add(this.res.getString("wf"));
			}
		} 
	catch (SQLException e) {	
		e.printStackTrace();
		lemResp.head_errMess = String.format("Error: %s: cannot convert all lemmata! returns after %d lemmata.\n", 
				func, this.nWfs);
		return lemResp;
		}

	lemResp.head_nWfs = this.nWfs;
	
	return lemResp;
	
	} // res2JSON
	
	
	/* res2TEXT():
	 * 
	 * transforms result of Lemma Query to a list of wordforms in Text/plain format.
	 * 28.01.20/FB
	 */
	
	public String res2TEXT()
	
	{
	final String
		func = "res2TEST";
	StringBuilder
		result;
	
	try {
		result = new StringBuilder();
		} 
	catch (Exception e) {
		GlemmServices.ferr.printf("Error: %s: cannot allocated memory!\n",  func); 
		e.printStackTrace();
		return String.format("Error: %s: cannot allocate memory for wordform list!\n", func);
		}
	
	// fill StringBuffer with list of wfs:
	try {
		for(this.nWfs=0; this.res.next(); this.nWfs++)
			{
			result.append(this.res.getString("wf") + "\n");
			}
		} 
	catch (SQLException e) {	
		e.printStackTrace();
		return String.format("Error: %s: cannot convert all lemmata! returns after %d lemmata.\n", 
				func, this.nWfs);
		}
	
	return result.toString();
	
	} // res2TEXT
	
	/*
	 * decompositionRulestoString:
	 * 
	 * transforms the GLEMM decomposition rules prop1+2 to a human
	 * readable string.
	 * See e.g. COSMAS II: src/util/list2sgml.c.
	 * 15.01.20/FB
	 * 21.10.21/FB C2-option 'Spec' is no Glemm option and thus not returned here.
	 */
	
	static String decompRulestoString(int prop1, int prop2)
	
	{
	if( prop1 == 0 && prop2 == 0)
		return "Flexionsformen";
	else if( prop1 == 1 && prop2 == 0)
		return "Komposita";
	else if( prop1 == 0 && prop2 == 1)
		return "sonst. Wortbidlungsformen";
	else if( prop1 == 1 && prop2 == 1)
		return "Komposita und sonst. Wortbildungsformen";
	else
		return "";
	
	} // decompositionRulestoString
	
	/*
	 * isWfSpecialCase
	 * Returns true, if wordform should be returned for GLEMM_OPT_SPEC,
	 * i.e. if it contains hyphens or special chars.
	 * 15.01.20/FB
	 */
	
	static boolean isWfSpecialCase(String wf)
	
	{
	final String
		func = "isWfSpecialCase";
	boolean
		isLorD;
	
	/* test: isLetterOrDigit works as expected: all chars from ISO8859-1 are identified as a letter,
	 * digits as digits, interpunction and other chars are identified as non-letter.
	 */
	/*
	char[]
		chars = new char[256];
	for(i=0; i<256; i++)
		chars[i] = (char)i;
	for(i=0; i<chars.length; i++)
		{
		type = Character.isLetterOrDigit(chars[i]);
		System.out.printf("Debug: %c : %s\n", chars[i], type == true ? "letter or digit" : "special char");
		}
	*/

	for(int i=0; i<wf.length(); i++)
		{
		isLorD = Character.isLetterOrDigit(wf.charAt(i));
		//System.out.printf("Debug: %c : %s\n", wf.charAt(i), isLorD == true ? "letter or digit" : "special char");
		if( !isLorD )
			{
			//System.out.printf("Debug: %s : %s spec=%b.\n", func, wf, true);
			return true; // wf contains at least one hyphen or special char.
			}
		}

	//System.out.printf("Debug: %s : %s spec=%b.\n", func, wf, false);
	
	return false;
	} // isWfSpecialCase
	
	
}

